//
// Created by zygzagz on 02.01.2021.
//

#ifndef ZAD2_VARIABLE_H
#define ZAD2_VARIABLE_H

#include <utility>
#include <map>

#include "../TypeDefs.h"
#include "../Info.h"

class Variable : std::enable_shared_from_this<Variable> {
public:
	Variable() = default;
	explicit Variable(VarInfoPtr info) : info(std::move(info)), constExpr(false), val(0) {};
	explicit Variable(int constVal) : type(new Int), constExpr(true), val(constVal) {};
	explicit Variable(std::string symbolName) {};
	VarInfoPtr info;
	TypePtr type;
	bool constExpr;
	int val;
	std::string name;
	// chcemy znać lokalizację w pamięci gdzie zapisać z powrotem i gdzie leży ogólnie
	// lok. może nie być, jeśli to tymczasowa
	// registers
	// is mem?
	// where in mem - id or offset
};

class VariableLocation {
public:
	VariableLocation() = default;
	VariableLocation(VariablePtr base, VariablePtr index, int multiplier, int offset)
		: base(std::move(base)), index(std::move(index)), multiplier(multiplier), offset(offset)
	{}
	VariablePtr base;
	VariablePtr index;
	int multiplier;
	int offset;

	string toString() const {
		std::string ret = "[" + base->name;
		if (multiplier || !(index->constExpr || !index->val)) ret += " + " + to_string(multiplier) + "*" + index->name;
		if (offset) ret += " + " + to_string(offset);
		return ret + "]";
	}
};


class VariableLayout {
public:
	map<VarInfoPtr, pair<VariablePtr, VariablePtr>> changes;
	void add(const VarInfoPtr& info) {
		if (!changes.count(info) && info->loc) {
			changes[info] = {info->loc, nullptr};
		}
	}

	void capture() {
		for (auto &p : changes) {
			p.second.second = p.first->loc;
		}
		for (auto it = changes.begin(); it != changes.end(); ) {
			auto pit = it;
			it++;
			if (pit->second.first == pit->second.second)
				changes.erase(pit);
		}
	}

	void revert() {
		for (const auto& p : changes) {
			VarInfoPtr info = p.first;
			info->loc = p.second.first;
		}
	}
};

#endif //ZAD2_VARIABLE_H
